;;; bench2.el --- benchmark read-extended-command -*- lexical-binding: t -*-

;; Load libraries with command-modes and completion-predicates.

(require 'bookmark)
(require 'dbus)
(require 'eww)
(require 'finder)
(require 'gnus-salt)
(require 'info)
(require 'man)
(require 'shr)
(require 'tramp-cmds)
(require 'tramp-crypt)

;; Custom read-extended-command-predicates.

(defun bench--safe (symbol buffer)
  (if (get symbol 'completion-predicate)
      (condition-case-unless-debug err
          (funcall (get symbol 'completion-predicate) symbol buffer)
        (error
         (message "command-completion-default-include-p: %s: %S" symbol err)
         nil))
    (or (null (command-modes symbol))
        (command-completion-using-modes-p symbol buffer))))

(defun bench--expensive (_symbol _buffer)
  (let ((i 1000))
    (dotimes (_ i)
      (setq i (1+ i)))
    i))

;; Custom read-extended-command.

(defun bench-read-extended-command ()
  (let ((buffer (current-buffer)))
    (minibuffer-with-setup-hook
        (lambda ()
          (add-hook 'post-self-insert-hook
                    (lambda ()
                      (setq execute-extended-command--last-typed
                            (minibuffer-contents)))
                    nil 'local)
          (setq-local minibuffer-default-add-function
	              (lambda ()
	                ;; Get a command name at point in the original buffer
	                ;; to propose it after M-n.
	                (let ((def
                               (with-current-buffer
			           (window-buffer (minibuffer-selected-window))
			         (and (commandp (function-called-at-point))
				      (format
                                       "%S" (function-called-at-point)))))
		              (all (sort (minibuffer-default-add-completions)
                                         #'string<)))
		          (if def
		              (cons def (delete def all))
		            all)))))
      ;; Read a string, completing from and restricting to the set of
      ;; all defined commands.  Don't provide any initial input.
      ;; Save the command read on the extended-command history list.
      (completing-read
       (concat (cond
	        ((eq current-prefix-arg '-) "- ")
	        ((and (consp current-prefix-arg)
		      (eq (car current-prefix-arg) 4)) "C-u ")
	        ((and (consp current-prefix-arg)
		      (integerp (car current-prefix-arg)))
	         (format "%d " (car current-prefix-arg)))
	        ((integerp current-prefix-arg)
	         (format "%d " current-prefix-arg)))
	       ;; This isn't strictly correct if `execute-extended-command'
	       ;; is bound to anything else (e.g. [menu]).
	       ;; It could use (key-description (this-single-command-keys)),
	       ;; but actually a prompt other than "M-x" would be confusing,
	       ;; because "M-x" is a well-known prompt to read a command
	       ;; and it serves as a shorthand for "Extended command: ".
               (if (memq 'shift (event-modifiers last-command-event))
	           "M-X "
	         "M-x "))
       (lambda (string pred action)
         (if (and suggest-key-bindings (eq action 'metadata))
	     '(metadata
	       (affixation-function . read-extended-command--affixation)
	       (category . command))
           (complete-with-action action obarray string pred)))
       (lambda (sym)
         (and (commandp sym)
              (or (null read-extended-command-predicate)
                  (and (functionp read-extended-command-predicate)
                       (condition-case-unless-debug err
                           (funcall read-extended-command-predicate sym buffer)
                         (error
                          (message
                           "command-completion-default-include-p: %s: %S"
                           sym err)))))))
       t nil 'extended-command-history))))

;; Benchmarks.

(with-current-buffer-window "*bench*" nil nil
  (pcase-dolist (`(,reps . ,keys)
                 '((10 . "\C-i\C-m")
                   (100 . "eww-\C-i\C-a\C-k\C-m")))
    (insert (if (bobp) "" ?\n) (key-description keys) ?\n)
    (setq keys (listify-key-sequence keys))
    (dotimes (_ 3)
      (insert ?\n)
      (dolist (fn '(read-extended-command bench-read-extended-command))
        (insert (symbol-name fn) ?\n)
        (dolist (pred '(bench--expensive
                        nil
                        command-completion-default-include-p
                        bench--safe))
          (let ((read-extended-command-predicate pred))
            (garbage-collect)
            (insert (apply #'format "  %-36s %f %d %f\n" pred
                           (benchmark-run reps
                             (setq unread-command-events keys)
                             (funcall fn))))))))))

;;; bench2.el ends here
